/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <libv4l2.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <glib.h>
#include "uv4l.h"

typedef void (*SetMenuItem) (V4LControlMenuItem*,
                             struct v4l2_querymenu*);

struct _V4LControlManager
{
        const char* device;
        int fd;
        V4LControl** controls;
        int size;
};

struct _V4LControl
{
        V4LControlManager* v4lmgr;
        int id;
        char* name;
        V4LControlType type;
        int64_t minimum;
        int64_t maximum;
        u_int64_t step;
        int64_t default_value;
        int menu_size;
        V4LControlMenuItem** menus;
};

struct _V4LControlMenuItem
{
        u_int32_t   cid;
        u_int32_t   index;
        const char* name;
        int64_t     value;
        bool        disabled;
};

static void
set_name(V4LControlMenuItem*    v4litem,
         struct v4l2_querymenu* qm)
{
        v4litem->name = g_strdup((char*)qm->name);
}

static void
set_value(V4LControlMenuItem*    v4litem,
          struct v4l2_querymenu* qm)
{
        v4litem->value = qm->value;
}

static V4LControlMenuItem*
disabled_menu_item(uint32_t    id,
                   u_int32_t   index) 
{
        V4LControlMenuItem* v4litem = g_new0(V4LControlMenuItem, 1);
        v4litem->cid = id;
        v4litem->index = index;
        v4litem->name = NULL;
        v4litem->value = INT64_MAX;
        v4litem->disabled = true;
        return v4litem;
}

static V4LControlMenuItem*
get_menu_item(int         fd,
              uint32_t    id,
              u_int32_t   index,
              SetMenuItem func)
{
        struct v4l2_querymenu qm;
        memset(&qm, 0, sizeof(qm));
        qm.id = id;
        qm.index = index;
        if (ioctl(fd, VIDIOC_QUERYMENU, &qm)) {
                return disabled_menu_item(id, index);
        }
        V4LControlMenuItem* v4litem = g_new0(V4LControlMenuItem, 1);
        v4litem->cid = qm.id;
        v4litem->index = qm.index;
        func(v4litem, &qm);
        return v4litem;
}

static void
v4l_control_set_menu(int         fd,
                     V4LControl* v4lctl,
                     SetMenuItem func)
{
        v4lctl->menu_size = v4lctl->maximum - v4lctl->minimum + 1;
        v4lctl->menus = g_new0(V4LControlMenuItem*, v4lctl->menu_size);
        for (int i = v4lctl->minimum; i <= v4lctl->maximum; i++) {
                V4LControlMenuItem* mi;
                mi = get_menu_item(fd, v4lctl->id, i, func);
                v4lctl->menus[i - v4lctl->minimum] = mi;
        }
}

V4LControlManager*
v4l_control_manager_new(const char* device)
{
        g_return_val_if_fail(device, NULL);
        int fd = open(device, O_RDONLY);
        if (fd < 0) {
                g_warning("Error opening device %s", device);
                return NULL;
        }
        V4LControlManager* v4lmgr = g_new0(V4LControlManager, 1);
        v4lmgr->device = g_strdup(device);
        v4lmgr->fd = fd;
        struct v4l2_query_ext_ctrl qc;
        qc.id = V4L2_CTRL_FLAG_NEXT_CTRL;
        GList* cl = NULL;
        while (!ioctl(fd, VIDIOC_QUERY_EXT_CTRL, &qc)) {
                V4LControl* v4lctl =
                        v4l_control_new(v4lmgr,
                                        qc.id,
                                        qc.name,
                                        qc.type,
                                        qc.minimum,
                                        qc.maximum,
                                        qc.step,
                                        qc.default_value);
                if (v4lctl->type == V4L_CONTROL_TYPE_MENU)
                        v4l_control_set_menu(fd, v4lctl, set_name);
                if (v4lctl->type == V4L_CONTROL_TYPE_INT_MENU)
                        v4l_control_set_menu(fd, v4lctl, set_value);
                cl = g_list_prepend(cl, v4lctl);
                v4lmgr->size++;
                qc.id |= V4L2_CTRL_FLAG_NEXT_CTRL;
        }
        cl = g_list_reverse(cl);
        v4lmgr->controls = g_new0(V4LControl*, v4lmgr->size);
        int i = 0;
        GList* tmp = cl;
        while (tmp != NULL) {
                v4lmgr->controls[i++] = tmp->data;
                tmp = tmp->next;
        }
        g_list_free(cl);

        return v4lmgr;
}

int
v4l_control_manager_get_size(V4LControlManager* v4lmgr)
{
        g_return_val_if_fail(v4lmgr, -1);
        return v4lmgr->size;
}

V4LControl*
v4l_control_manager_nth_control(V4LControlManager* v4lmgr,
                                int                nth)
{
        g_return_val_if_fail(v4lmgr, NULL);
        return v4lmgr->controls[nth];
}

void
v4l_control_manager_free(V4LControlManager* v4lmgr)
{
        g_return_if_fail(v4lmgr);
        if (close(v4lmgr->fd))
                g_warning("Error closing device %s", v4lmgr->device);
        for (int i = 0; i < v4lmgr->size; i++)
                v4l_control_free(v4lmgr->controls[i]);
        g_free(v4lmgr->controls);
        g_free(v4lmgr);
}

V4LControl*
v4l_control_new(V4LControlManager* v4lmgr,
                int                id,
                char*              name,
                V4LControlType     type,
                int64_t            minimum,
                int64_t            maximum,
                u_int64_t          step,
                int64_t            default_value)
{
        g_return_val_if_fail(v4lmgr && id >= V4L2_CID_BASE && name, NULL);
        V4LControl* v4lctl = g_new0(V4LControl, 1);
        v4lctl->v4lmgr = v4lmgr;
        v4lctl->id = id;
        v4lctl->name = g_strdup(name);
        v4lctl->type = type;
        v4lctl->minimum = minimum;
        v4lctl->maximum = maximum;
        v4lctl->step = step;
        v4lctl->default_value = default_value;

        return v4lctl;
}

int
v4l_control_get_id(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, -1);
        return v4lctl->id;
}

char*
v4l_control_get_name(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, NULL);
        return v4lctl->name;
}

V4LControlType
v4l_control_get_type(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, V4L_CONTROL_TYPE_INVALID);
        return v4lctl->type;
}

int64_t
v4l_control_get_minimum(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, INT64_MAX);
        return v4lctl->minimum;
}

int64_t
v4l_control_get_maximum(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, INT64_MIN);
        return v4lctl->maximum;
}

u_int64_t
v4l_control_get_step(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, 0);
        return v4lctl->step;
}

int64_t
v4l_control_get_default_value(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, INT64_MAX);
        return v4lctl->default_value;
}

int64_t
v4l_control_get_value(V4LControl* v4lctl) 
{
        g_return_val_if_fail(v4lctl, INT64_MAX);
        struct v4l2_control c;
        memset(&c, 0, sizeof(c));
        c.id = v4lctl->id;
        if (ioctl(v4lctl->v4lmgr->fd, VIDIOC_G_CTRL, &c))
                g_warning("VIDIOC_G_CTRL failed for %s", v4lctl->name);
        return c.value;
}

void
v4l_control_set_value(V4LControl* v4lctl,
                      int64_t     value)
{
        g_return_if_fail(v4lctl);
        struct v4l2_control c;
        memset(&c, 0, sizeof(c));
        c.id = v4lctl->id;
        c.value = value;
        if (ioctl(v4lctl->v4lmgr->fd, VIDIOC_S_CTRL, &c))
                g_warning("VIDIOC_S_CTRL failed for %s", v4lctl->name);
}

int
v4l_control_get_menu_size(V4LControl* v4lctl)
{
        g_return_val_if_fail(v4lctl, -1);
        return v4lctl->menu_size;
}

V4LControlMenuItem*
v4l_control_nth_menu_item(V4LControl* v4lctl,
                          int         n)
{
        g_return_val_if_fail(v4lctl, NULL);
        return v4lctl->menus[n];
}

void
v4l_control_free(V4LControl* v4lctl)
{
        if (!v4lctl)
                return;
        g_free(v4lctl->name);
        g_free(v4lctl);
}

u_int32_t
v4l_control_menu_item_get_cid(V4LControlMenuItem *v4lcmi)
{
        g_return_val_if_fail(v4lcmi, INT_MAX);
        return v4lcmi->cid;
}

u_int32_t
v4l_control_menu_item_get_index(V4LControlMenuItem *v4lcmi)
{
        g_return_val_if_fail(v4lcmi, INT_MAX);
        return v4lcmi->index;
}

const char*
v4l_control_menu_item_get_name(V4LControlMenuItem *v4lcmi)
{
        g_return_val_if_fail(v4lcmi, NULL);
        return v4lcmi->name;
}

int64_t
v4l_control_menu_item_get_value (V4LControlMenuItem *v4lcmi)
{
        g_return_val_if_fail(v4lcmi, INT64_MAX);
        return v4lcmi->value;
}

bool
v4l_control_menu_item_is_disabled(V4LControlMenuItem *v4lcmi)
{
        g_return_val_if_fail(v4lcmi, true);
        return v4lcmi->disabled;
}
