/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class to handle the pipeline.
     *
     * Raw idle pipeline:
     *
     *{{{
     *                          ┌─────────────────────┐
     *                          |      video-src      |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |    video-filter     |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-queue-gtk   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-convert-gtk  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-text-record  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-text-info   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          | video-text-message  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-gtk-sink    |
     *                          └─────────────────────┘
     * }}}
     *
     * Raw recording pipeline:
     *
     * {{{
     *                                                 ┌─────────────────────┐
     *                                                 |      audio-src      |
     *                                                 └──────────┬──────────┘
     *                                                            │
     *             ┌─────────────────────┐             ┌──────────┴──────────┐
     *             |      video-src      |             │    audio-filter     │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |    video-filter     |             │  audio-queue-rate   │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   |
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |      video-tee      │             |     audio-rate      │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *            ┌───────────┴───────────┐                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |   video-queue-gtk   │ │  video-queue-file   │ │    audio-convert    │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       │                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |  video-convert-gtk  | | video-convert-file  | │   audio-encoder     │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       │                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |  video-text-record  | │ video-h264-encoder  │ │ audio-queue-mp4mux  │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       └───────────┬───────────┘
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * |   video-text-info   |             |     file-mp4mux     |
     * └──────────┬──────────┘             └──────────┬──────────┘
     *            │                                   │
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * | video-text-message  |             |      file-sink      │
     * └──────────┬──────────┘             └─────────────────────┘
     *            │
     * ┌──────────┴──────────┐
     * |   video-gtk-sink    |
     * └─────────────────────┘
     * }}}
     *
     * H.264 idle pipeline:
     *
     *{{{
     *                          ┌─────────────────────┐
     *                          |      video-src      |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |    video-filter     |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-queue-parse  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-h264-parse   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-queue-gtk   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          | video-h264-decoder  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          | video-queue-convert |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-convert-gtk  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |  video-text-record  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-text-info   |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          | video-text-message  |
     *                          └──────────┬──────────┘
     *                                     │
     *                          ┌──────────┴──────────┐
     *                          |   video-gtk-sink    |
     *                          └─────────────────────┘
     * }}}
     *
     * H.264 recording pipeline:
     *
     * {{{
     *
     *                                                 ┌─────────────────────┐
     *                                                 |      audio-src      |
     *                                                 └──────────┬──────────┘
     *                                                            │
     *             ┌─────────────────────┐             ┌──────────┴──────────┐
     *             |      video-src      |             |     audio-filter    |
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |    video-filter     |             │  audio-queue-rate   │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |  video-queue-parse  |             │     audio-rate      │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   |
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |  video-h264-parse   |             |    audio-convert    │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *                        │                                   │
     *             ┌──────────┴──────────┐             ┌──────────┴──────────┐
     *             |      video-tee      |             │    audio-encoder    │
     *             └──────────┬──────────┘             └──────────┬──────────┘
     *            ┌───────────┴───────────┐                       │
     * ┌──────────┴──────────┐ ┌──────────┴──────────┐ ┌──────────┴──────────┐
     * |   video-queue-gtk   │ │  video-queue-file   │ │ audio-queue-mp4mux  │
     * └──────────┬──────────┘ └──────────┬──────────┘ └──────────┬──────────┘
     *            │                       └───────────┬───────────┘
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * | video-h264-decoder  |             |     file-mp4mux     |
     * └──────────┬──────────┘             └──────────┬──────────┘
     *            │                                   │
     * ┌──────────┴──────────┐             ┌──────────┴──────────┐
     * | video-queue-convert |             │      file-sink      │
     * └──────────┬──────────┘             └─────────────────────┘
     *            │
     * ┌──────────┴──────────┐
     * |  video-convert-gtk  |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * |  video-text-record  |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * |   video-text-info   |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * | video-text-message  |
     * └──────────┬──────────┘
     *            │
     * ┌──────────┴──────────┐
     * |   video-gtk-sink    |
     * └─────────────────────┘
     * }}}
     *
     */
    public class Camera : GLib.Object {

        private class Message : GLib.Object {
            public double timestamp;
            public string text;

            public Message(string text) {
                timestamp = Timestamp.get_timestamp();
                this.text = text;
            }
        }

        private Gtk.Widget _widget;
        public unowned Gtk.Widget widget {
            get {
                video_gtk_sink.get("widget", out _widget);
                return _widget;
            }
        }

        public bool saving { public get; private set; }

        /* The video sources. */
        private VideoSources video_sources;
        private VideoSource video_source;
        private VideoSource new_video_source;
        private VideoSetup video_setup;
        private VideoSetup new_video_setup;

        /* Pipeline and elements. */
        protected Gst.Pipeline pipeline;
        private Gst.Bus bus;

        /* Raw pipeline. */
        private Gst.Element video_src;
        private Gst.Element video_filter;
        private Gst.Element video_queue_convert;
        private Gst.Element video_queue_gtk;
        private Gst.Element video_convert_gtk;
        private Gst.Element video_text_record;
        private Gst.Element video_text_info;
        private Gst.Element video_text_message;
        private Gst.Element video_gtk_sink;

        /* H.264 elements. */
        private Gst.Element video_queue_parse;
        private Gst.Element video_h264_parse;
        private Gst.Element video_h264_decoder;

        /* Audio pipeline. */
        private Gst.Element audio_src;
        private Gst.Element audio_filter;
        private Gst.Element audio_queue_rate;
        private Gst.Element audio_rate;
        private Gst.Element audio_convert;
        private Gst.Element audio_encoder;
        private Gst.Element audio_queue_mp4mux;

        /* File sink branch. */
        private Gst.Element video_tee;
        private Gst.Element video_queue_file;
        private Gst.Element video_convert_file;
        private Gst.Element video_h264_encoder;
        private Gst.Element file_mp4mux;

        /* File sink*/
        private Gst.Element file_sink;

        /* Temporary filename . */
        private string temporary_fn;

        private Gee.Queue<Message> messages;

        /* Timestamps. */
        private double start_saving_ts;
        private double last_saving_ts;
        private double message_queue_ts;

        private ulong block_probe_id;
        private int dot_counter;
        private bool waiting_keyframe;
        private bool video_saving;
        private bool show_info;
        private int debug_counter;
        private bool last_change_done;
        private bool queuing_messages;

        public Camera(VideoSources video_sources) {
            this.video_sources = video_sources;
            messages = new Gee.ArrayQueue<Message>();
            video_sources.notify["video-source"].connect(video_source_changed);
            video_source = video_sources.video_source;
            video_setup = video_source.setups.first();

            pipeline = new Gst.Pipeline("queso");

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                raw_idle_pipeline();
                break;
            case VideoSetup.H264:
                h264_idle_pipeline();
                break;
            }

            bus = pipeline.get_bus();
            bus.add_watch(GLib.Priority.DEFAULT, pipeline_bus_watch);
            last_change_done = true;
        }

        /**
         * Starts the camera.
         */
        public void start() {
            GLib.Idle.add(start_camera);
            queuing_messages = true;
            message_queue_ts = 0.0;
            GLib.Idle.add(handle_message_queue);
        }

        /**
         * Stops the camera.
         */
        public void stop() {
            GLib.Idle.add(stop_camera);
            queuing_messages = false;
        }

        /**
         * Cleans the camera and stops the pipeline.
         */
        public void clean() {
            pipeline.send_event(new Gst.Event.eos());
            if (file_sink != null)
                file_sink.set_state(Gst.State.NULL);
            if (file_mp4mux != null)
                file_mp4mux.set_state(Gst.State.NULL);
            if (saving) {
                var destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                destination_fn = temporary_fn = null;
            }
        }

        /**
         * Toggles saving the camera video.
         */
        public void toggle_save() {
            saving = !saving;
            if (saving) {
                start_saving();
            } else {
                stop_saving();
            }
        }

        /**
         * Toggles the device information in the feed.
         */
        public void toggle_info() {
            show_info = !show_info;
            if (show_info) {
                video_text_info.set("text", get_info());
            } else {
                video_text_info.set("text", "");
            }
        }

        /**
         * Stores a dot file with the current pipeline, in the directory defined
         * by ''GST_DEBUG_DUMP_DOT_DIR'', with the name
         * ''queso-pipeline-XXX.dot.
         */
        public void debug() {
            if (GLib.Environment.get_variable("GST_DEBUG_DUMP_DOT_DIR") == null)
                return;
            var output = "queso-pipeline-%03d".printf(debug_counter++);
            Gst.Debug.bin_to_dot_file(pipeline, Gst.DebugGraphDetails.ALL,
                                      output);
            add_message("Dot graph saved in %s", output);
        }

        /* Pipeline bus watch. */
        private bool
        pipeline_bus_watch(Gst.Bus bus, Gst.Message message) {
            switch (message.type) {
            case Gst.MessageType.ERROR:
                debug();
                GLib.Error error;
                string debug;
                message.parse_error(out error, out debug);
                GLib.warning("ERROR: " + error.message);
                return GLib.Source.CONTINUE;
            case Gst.MessageType.WARNING:
                debug();
                GLib.Error error;
                string debug;
                message.parse_warning(out error, out debug);
                GLib.warning("WARNING: " + error.message);
                return GLib.Source.CONTINUE;
            case Gst.MessageType.INFO:
                debug();
                GLib.Error error;
                string debug;
                message.parse_info(out error, out debug);
                GLib.warning("INFO: " + error.message);
                return GLib.Source.CONTINUE;
            case Gst.MessageType.ELEMENT:
                unowned Gst.Structure structure = message.get_structure();
                if (!structure.has_name("GstBinForwarded"))
                    return GLib.Source.CONTINUE;
                Gst.Message forward_message = null;
                structure.get("message", typeof(Gst.Message),
                              out forward_message);
                if (forward_message.type != Gst.MessageType.EOS)
                    return GLib.Source.CONTINUE;

                switch (video_setup.media_type) {
                case VideoSetup.RAW:
                    raw_eos();
                    break;
                case VideoSetup.H264:
                    h264_eos();
                    break;
                }

                var destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                temporary_fn = destination_fn = null;
                return GLib.Source.CONTINUE;
            default:
                return GLib.Source.CONTINUE;
            }
        }

        /* Closes the file for a raw saving. */
        private void raw_eos() {
            file_mp4mux.set_state(Gst.State.NULL);
            file_sink.set_state(Gst.State.NULL);
            file_sink.set("location", "/dev/null");
            video_h264_encoder.set_state(Gst.State.NULL);
            video_convert_file.set_state(Gst.State.NULL);
            video_queue_file.set_state(Gst.State.NULL);
            video_tee.set_state(Gst.State.NULL);

            video_text_record.set("text", "");
            add_message("Recording stopped");

            audio_queue_mp4mux.set_state(Gst.State.NULL);
            audio_encoder.set_state(Gst.State.NULL);
            audio_convert.set_state(Gst.State.NULL);
            audio_rate.set_state(Gst.State.NULL);
            audio_queue_rate.set_state(Gst.State.NULL);
            audio_filter.set_state(Gst.State.NULL);
            audio_src.set_state(Gst.State.NULL);

            pipeline.remove_many(file_sink,
                                 file_mp4mux,
                                 video_h264_encoder,
                                 video_convert_file,
                                 video_queue_file,
                                 video_tee,
                                 audio_queue_mp4mux,
                                 audio_encoder,
                                 audio_convert,
                                 audio_rate,
                                 audio_queue_rate,
                                 audio_filter,
                                 audio_src);

            video_tee = null;
            video_queue_file = null;
            video_convert_file = null;
            video_h264_encoder = null;
            file_mp4mux = null;
            file_sink = null;

            audio_src = null;
            audio_filter = null;
            audio_queue_rate = null;
            audio_rate = null;
            audio_convert = null;
            audio_encoder = null;
            audio_queue_mp4mux = null;

            var src_pad = video_filter.get_static_pad("src");
            var sink_pad = video_queue_gtk.get_static_pad("sink");
            src_pad.link(sink_pad);
        }

        /* Closes the file for a H264 saving. */
        private void h264_eos() {
            file_mp4mux.set_state(Gst.State.NULL);
            file_sink.set_state(Gst.State.NULL);
            file_sink.set("location", "/dev/null");
            video_queue_file.set_state(Gst.State.NULL);
            video_tee.set_state(Gst.State.NULL);

            video_text_record.set("text", "");
            add_message("Recording stopped");

            audio_queue_mp4mux.set_state(Gst.State.NULL);
            audio_encoder.set_state(Gst.State.NULL);
            audio_convert.set_state(Gst.State.NULL);
            audio_rate.set_state(Gst.State.NULL);
            audio_queue_rate.set_state(Gst.State.NULL);
            audio_filter.set_state(Gst.State.NULL);
            audio_src.set_state(Gst.State.NULL);

            pipeline.remove_many(file_sink,
                                 file_mp4mux,
                                 video_queue_file,
                                 video_tee,
                                 audio_queue_mp4mux,
                                 audio_encoder,
                                 audio_convert,
                                 audio_rate,
                                 audio_queue_rate,
                                 audio_filter,
                                 audio_src);

            video_tee = null;
            video_queue_file = null;
            file_mp4mux = null;
            file_sink = null;

            audio_src = null;
            audio_filter = null;
            audio_queue_rate = null;
            audio_rate = null;
            audio_convert = null;
            audio_encoder = null;
            audio_queue_mp4mux = null;

            var src_pad = video_h264_parse.get_static_pad("src");
            var sink_pad = video_queue_gtk.get_static_pad("sink");
            src_pad.link(sink_pad);
        }

        /* Starts the camera asynchronously. */
        private bool start_camera() {
            return change_pipeline_state(Gst.State.PLAYING);
        }

        /* Stops the camera asynchronously. */
        private bool stop_camera() {
            return change_pipeline_state(Gst.State.PAUSED);
        }

        /* Changes the pipeline state, returning until it's changed. */
        private bool change_pipeline_state(Gst.State new_state) {
            pipeline.set_state(new_state);
            Gst.State state = Gst.State.NULL;
            Gst.State pending;

            Gst.StateChangeReturn r;

            do {
                r = pipeline.get_state(out state, out pending, 100);
                if (r == Gst.StateChangeReturn.FAILURE) {
                    return false;
                }
                Thread.usleep(10);
            } while (state != new_state);

            return true;
        }

        /* Sets the raw idle pipeline. */
        private void raw_idle_pipeline() {
            video_src = video_source.create_element("video-src");
            video_filter = Gst.ElementFactory.make(
                "capsfilter", "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_gtk = Gst.ElementFactory.make(
                "queue", "video-queue-gtk");
            video_convert_gtk = Gst.ElementFactory.make(
                "videoconvert", "video-convert-gtk");
            video_text_record = Gst.ElementFactory.make(
                "textoverlay", "video-text-record");
            video_text_record.set("font-desc", "sans 14");
            video_text_record.set("halignment", 2);
            video_text_record.set("valignment", 2);
            video_text_info = Gst.ElementFactory.make(
                "textoverlay", "video-text-info");
            video_text_info.set("font-desc", "sans 10");
            video_text_info.set("halignment", 0);
            video_text_info.set("valignment", 2);
            video_text_info.set("line-alignment", 0);
            video_text_message = Gst.ElementFactory.make(
                "textoverlay", "video-text-message");
            video_text_message.set("font-desc", "sans 10");
            video_text_message.set("halignment", 0);
            video_text_message.set("valignment", 0);
            video_text_message.set("line-alignment", 0);
            video_gtk_sink = Gst.ElementFactory.make(
                "gtksink", "video-gtk-sink");

            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_gtk,
                              video_convert_gtk,
                              video_text_record,
                              video_text_info,
                              video_text_message,
                              video_gtk_sink);
            video_src.link_many(video_filter,
                                video_queue_gtk,
                                video_convert_gtk,
                                video_text_record,
                                video_text_info,
                                video_text_message,
                                video_gtk_sink);
        }

        /* Sets the H264 idle pipeline. */
        private void h264_idle_pipeline() {
            video_src = video_source.create_element("video-src");
            video_filter = Gst.ElementFactory.make(
                "capsfilter", "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_parse = Gst.ElementFactory.make(
                "queue", "video-queue-parse");
            video_queue_parse.set("max-size-time", 3 * Gst.SECOND);
            video_queue_parse.set("max-size-bytes", 0);
            video_queue_parse.set("max-size-buffers", 0);
            video_queue_parse.set("leaky", 1);
            video_queue_parse.set("flush-on-eos", true);
            video_h264_parse = Gst.ElementFactory.make(
                "h264parse", "video-h264-parse");
            video_h264_parse.set("config-interval", 1);
            video_h264_parse.set("disable-passthrough", true);
            video_queue_gtk = Gst.ElementFactory.make(
                "queue", "video-queue-gtk");
            video_h264_decoder = Gst.ElementFactory.make(
                "avdec_h264", "video-h264-decoder");
            video_queue_convert = Gst.ElementFactory.make(
                "queue", "video-queue-convert");
            video_convert_gtk = Gst.ElementFactory.make(
                "videoconvert", "video-convert-gtk");
            video_text_record = Gst.ElementFactory.make(
                "textoverlay", "video-text-record");
            video_text_record.set("font-desc", "sans 14");
            video_text_record.set("halignment", 2);
            video_text_record.set("valignment", 2);
            video_text_info = Gst.ElementFactory.make(
                "textoverlay", "video-text-info");
            video_text_info.set("font-desc", "sans 10");
            video_text_info.set("halignment", 0);
            video_text_info.set("valignment", 2);
            video_text_info.set("line-alignment", 0);
            video_text_message = Gst.ElementFactory.make(
                "textoverlay", "video-text-message");
            video_text_message.set("font-desc", "sans 10");
            video_text_message.set("halignment", 0);
            video_text_message.set("valignment", 0);
            video_text_message.set("line-alignment", 0);
            video_gtk_sink = Gst.ElementFactory.make(
                "gtksink", "video-gtk-sink");

            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_parse,
                              video_h264_parse,
                              video_queue_gtk,
                              video_h264_decoder,
                              video_queue_convert,
                              video_convert_gtk,
                              video_text_record,
                              video_text_info,
                              video_text_message,
                              video_gtk_sink);
            video_src.link_many(video_filter,
                                video_queue_parse,
                                video_h264_parse,
                                video_queue_gtk,
                                video_h264_decoder,
                                video_queue_convert,
                                video_convert_gtk,
                                video_text_record,
                                video_text_info,
                                video_text_message,
                                video_gtk_sink);
        }

        /* Starts saving. */
        private void start_saving() {
            dot_counter = 0;
            saving = true;
            waiting_keyframe = true;
            video_text_record.set("text", "Please wait for keyframe");
            video_text_record.set("halignment", 1);
            video_text_record.set("valignment", 1);
            GLib.Idle.add(update_saving_counter);
            Gst.Pad pad = null;
            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                pad = video_queue_gtk.get_static_pad("src");
                break;
            case VideoSetup.H264:
                pad = video_queue_parse.get_static_pad("src");
                break;
            }
            pad.add_probe(Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                          start_saving_block_probe);
        }

        /* Stops saving. */
        private void stop_saving() {
            var pad = file_mp4mux.get_static_pad("src");
            block_probe_id = pad.add_probe(Gst.PadProbeType.BLOCK |
                                           Gst.PadProbeType.BUFFER,
                                           stop_saving_block_probe);

            new Thread<void*>("eos-push-thread", push_eos_thread);
        }

        /* Change video source block probe. */
        private Gst.PadProbeReturn
        start_saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            pad.remove_probe(info.id);

            switch (video_setup.media_type) {
            case VideoSetup.RAW:
                raw_add_saving_pipeline();
                break;
            case VideoSetup.H264:
                h264_add_saving_pipeline();
                break;
            }

            return Gst.PadProbeReturn.OK;
        }

        private void raw_add_saving_pipeline() {
            var src_pad = video_filter.get_static_pad("src");
            var sink_pad = video_queue_gtk.get_static_pad("sink");
            src_pad.unlink(sink_pad);

            audio_src = Gst.ElementFactory.make("pulsesrc", "audio-src");
            audio_filter = Gst.ElementFactory.make("capsfilter",
                                                   "audio-filter");
            var caps = new Gst.Caps.empty_simple("audio/x-raw");
            caps.set_simple("channels", typeof(int), 2);
            audio_filter.set("caps", caps);
            audio_queue_rate = Gst.ElementFactory.make("queue",
                                                       "audio-queue-rate");
            audio_rate = Gst.ElementFactory.make("audiorate", "audio-rate");
            audio_convert = Gst.ElementFactory.make("audioconvert",
                                                    "audio-convert");
            audio_encoder = Gst.ElementFactory.make("avenc_aac",
                                                    "audio_encoder");
            audio_encoder.set("perfect-timestamp", true);
            audio_queue_mp4mux = Gst.ElementFactory.make("queue",
                                                         "audio_queue_mp4mux");
            audio_queue_mp4mux.set("max-size-time", 3 * Gst.SECOND);
            audio_queue_mp4mux.set("max-size-bytes", 0);
            audio_queue_mp4mux.set("max-size-buffers", 0);
            audio_queue_mp4mux.set("leaky", 2);

            video_tee = Gst.ElementFactory.make("tee", "video-tee");
            video_queue_file = Gst.ElementFactory.make("queue",
                                                       "video-queue-file");
            video_queue_file.set("max-size-time", Gst.SECOND);
            video_queue_file.set("max-size-bytes", 0);
            video_queue_file.set("max-size-buffers", 0);
            video_queue_file.set("leaky", 2);
            video_convert_file = Gst.ElementFactory.make("videoconvert",
                                                         "video-convert-file");
            video_h264_encoder = Gst.ElementFactory.make("x264enc",
                                                         "video-h264-encoder");
            video_h264_encoder.set("tune", 0x4);
            video_h264_encoder.set("speed-preset", 1);
            file_mp4mux = Gst.ElementFactory.make("mpegtsmux",
                                                  "file-mp4mux");
            file_sink = Gst.ElementFactory.make("filesink", "file-sink");
            file_sink.set("qos", true);
            file_sink.set("sync", true);
            file_sink.set("async", false);
            temporary_fn = random_filename();
            file_sink.set("location", temporary_fn);

            pipeline.add_many(audio_src,
                              audio_filter,
                              audio_queue_rate,
                              audio_rate,
                              audio_convert,
                              audio_encoder,
                              audio_queue_mp4mux,
                              video_tee,
                              video_queue_file,
                              video_convert_file,
                              video_h264_encoder,
                              file_mp4mux,
                              file_sink);
            video_filter.link(video_tee);

            audio_src.link_many(audio_filter,
                                audio_queue_rate,
                                audio_rate,
                                audio_convert,
                                audio_encoder,
                                audio_queue_mp4mux,
                                file_mp4mux);

            video_tee.link(video_queue_gtk);
            video_tee.link_many(video_queue_file,
                                video_convert_file,
                                video_h264_encoder,
                                file_mp4mux,
                                file_sink);

            video_saving = false;
            var pad = video_queue_file.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BUFFER, analyze_video_buffer_probe);
            pad = audio_queue_mp4mux.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BUFFER, analyze_audio_buffer_probe);

            audio_src.set_state(Gst.State.PLAYING);
            audio_filter.set_state(Gst.State.PLAYING);
            audio_queue_rate.set_state(Gst.State.PLAYING);
            audio_rate.set_state(Gst.State.PLAYING);
            audio_convert.set_state(Gst.State.PLAYING);
            audio_encoder.set_state(Gst.State.PLAYING);
            audio_queue_mp4mux.set_state(Gst.State.PLAYING);
            video_tee.set_state(Gst.State.PLAYING);
            video_queue_file.set_state(Gst.State.PLAYING);
            video_convert_file.set_state(Gst.State.PLAYING);
            video_h264_encoder.set_state(Gst.State.PLAYING);
            file_mp4mux.set_state(Gst.State.PLAYING);
            file_sink.set_state(Gst.State.PLAYING);
        }

        private void h264_add_saving_pipeline() {
            var src_pad = video_h264_parse.get_static_pad("src");
            var sink_pad = video_queue_gtk.get_static_pad("sink");
            src_pad.unlink(sink_pad);

            audio_src = Gst.ElementFactory.make("pulsesrc", "audio-src");
            audio_filter = Gst.ElementFactory.make("capsfilter",
                                                   "audio-filter");
            var caps = new Gst.Caps.empty_simple("audio/x-raw");
            caps.set_simple("channels", typeof(int), 2);
            audio_filter.set("caps", caps);
            audio_queue_rate = Gst.ElementFactory.make("queue",
                                                       "audio-queue-rate");
            audio_rate = Gst.ElementFactory.make("audiorate", "audio-rate");
            audio_convert = Gst.ElementFactory.make("audioconvert",
                                                    "audio-convert");
            audio_encoder = Gst.ElementFactory.make("avenc_aac",
                                                    "audio_encoder");
            audio_encoder.set("perfect-timestamp", true);
            audio_queue_mp4mux = Gst.ElementFactory.make("queue",
                                                         "audio_queue_mp4mux");
            audio_queue_mp4mux.set("max-size-time", 3 * Gst.SECOND);
            audio_queue_mp4mux.set("max-size-bytes", 0);
            audio_queue_mp4mux.set("max-size-buffers", 0);
            audio_queue_mp4mux.set("leaky", 2);

            video_tee = Gst.ElementFactory.make("tee", "video-tee");
            video_queue_file = Gst.ElementFactory.make("queue",
                                                       "video-queue-file");
            video_queue_file.set("max-size-time", 3 * Gst.SECOND);
            video_queue_file.set("max-size-bytes", 0);
            video_queue_file.set("max-size-buffers", 0);
            video_queue_file.set("leaky", 2);
            file_mp4mux = Gst.ElementFactory.make("mpegtsmux",
                                                  "file-mp4mux");
            file_sink = Gst.ElementFactory.make("filesink", "file-sink");
            file_sink.set("qos", true);
            file_sink.set("sync", true);
            file_sink.set("async", false);
            temporary_fn = random_filename();
            file_sink.set("location", temporary_fn);

            pipeline.add_many(audio_src,
                              audio_filter,
                              audio_queue_rate,
                              audio_rate,
                              audio_convert,
                              audio_encoder,
                              audio_queue_mp4mux,
                              video_tee,
                              video_queue_file,
                              file_mp4mux,
                              file_sink);
            video_h264_parse.link(video_tee);

            audio_src.link_many(audio_filter,
                                audio_queue_rate,
                                audio_rate,
                                audio_convert,
                                audio_encoder,
                                audio_queue_mp4mux,
                                file_mp4mux);

            video_tee.link(video_queue_gtk);
            video_tee.link_many(video_queue_file,
                                file_mp4mux,
                                file_sink);

            video_saving = false;
            var pad = video_queue_file.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BUFFER, analyze_video_buffer_probe);
            pad = audio_queue_mp4mux.get_static_pad("src");
            pad.add_probe(Gst.PadProbeType.BUFFER, analyze_audio_buffer_probe);

            audio_src.set_state(Gst.State.PLAYING);
            audio_filter.set_state(Gst.State.PLAYING);
            audio_queue_rate.set_state(Gst.State.PLAYING);
            audio_rate.set_state(Gst.State.PLAYING);
            audio_convert.set_state(Gst.State.PLAYING);
            audio_encoder.set_state(Gst.State.PLAYING);
            audio_queue_mp4mux.set_state(Gst.State.PLAYING);
            video_tee.set_state(Gst.State.PLAYING);
            video_queue_file.set_state(Gst.State.PLAYING);
            file_mp4mux.set_state(Gst.State.PLAYING);
            file_sink.set_state(Gst.State.PLAYING);
        }

        private Gst.PadProbeReturn
        analyze_video_buffer_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            Gst.Buffer buffer = (Gst.Buffer)info.data;
            var flags = buffer.get_flags();
            if ((flags & Gst.BufferFlags.DELTA_UNIT) == 0) {
                pad.remove_probe(info.id);
                video_saving = true;
                waiting_keyframe = false;
                start_saving_ts = Timestamp.get_timestamp();
                video_text_record.set("text", "•REC 00:00:00");
                video_text_record.set("halignment", 2);
                video_text_record.set("valignment", 2);
                add_message("Recording started");
                return Gst.PadProbeReturn.OK;
            }
            return Gst.PadProbeReturn.DROP;
        }

        private Gst.PadProbeReturn
        analyze_audio_buffer_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            if (!video_saving)
                return Gst.PadProbeReturn.DROP;
            return Gst.PadProbeReturn.OK;
        }


        /* Saving block probe. */
        private Gst.PadProbeReturn
        stop_saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            return Gst.PadProbeReturn.OK;
        }

        private void video_source_changed() {
            new_video_source = video_sources.video_source;
            new_video_setup = new_video_source.setups.first();
            if (video_setup.media_type == VideoSetup.RAW &&
                new_video_setup.media_type == VideoSetup.H264) {
                var pad = video_queue_gtk.get_static_pad("src");
                pad.add_probe(Gst.PadProbeType.BLOCK, raw_to_h264_block_probe);
            }
            if (video_setup.media_type == VideoSetup.H264 &&
                new_video_setup.media_type == VideoSetup.RAW) {
                var pad = video_queue_convert.get_static_pad("src");
                pad.add_probe(Gst.PadProbeType.BLOCK, h264_to_raw_block_probe);
            }
        }

        private Gst.PadProbeReturn
        raw_to_h264_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            pad.remove_probe(info.id);
            video_src.set_state(Gst.State.NULL);
            video_filter.set_state(Gst.State.NULL);
            video_queue_gtk.set_state(Gst.State.NULL);

            video_queue_gtk.unlink(video_convert_gtk);
            pipeline.remove_many(video_src,
                                 video_filter,
                                 video_queue_gtk);

            video_source = new_video_source;
            video_setup = video_source.setups.first();
            new_video_source = null;

            video_src = video_source.create_element("video-src");
            video_filter = Gst.ElementFactory.make(
                "capsfilter", "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_parse = Gst.ElementFactory.make(
                "queue", "video-queue-parse");
            video_queue_parse.set("max-size-time", 3 * Gst.SECOND);
            video_queue_parse.set("max-size-bytes", 0);
            video_queue_parse.set("max-size-buffers", 0);
            video_queue_parse.set("leaky", 1);
            video_queue_parse.set("flush-on-eos", true);
            video_h264_parse = Gst.ElementFactory.make(
                "h264parse", "video-h264-parse");
            video_h264_parse.set("config-interval", 1);
            video_h264_parse.set("disable-passthrough", true);
            video_queue_gtk = Gst.ElementFactory.make(
                "queue", "video-queue-gtk");
            video_h264_decoder = Gst.ElementFactory.make(
                "avdec_h264", "video-h264-decoder");
            video_queue_convert = Gst.ElementFactory.make(
                "queue", "video-queue-convert");

            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_parse,
                              video_h264_parse,
                              video_queue_gtk,
                              video_h264_decoder,
                              video_queue_convert);
            video_src.link_many(video_filter,
                                video_queue_parse,
                                video_h264_parse,
                                video_queue_gtk,
                                video_h264_decoder,
                                video_queue_convert,
                                video_convert_gtk);

            video_src.set_state(Gst.State.PLAYING);
            video_filter.set_state(Gst.State.PLAYING);
            video_queue_parse.set_state(Gst.State.PLAYING);
            video_h264_parse.set_state(Gst.State.PLAYING);
            video_queue_gtk.set_state(Gst.State.PLAYING);
            video_h264_decoder.set_state(Gst.State.PLAYING);
            video_queue_convert.set_state(Gst.State.PLAYING);

            return Gst.PadProbeReturn.DROP;
        }

        private Gst.PadProbeReturn
        h264_to_raw_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            pad.remove_probe(info.id);
            video_src.set_state(Gst.State.NULL);
            video_filter.set_state(Gst.State.NULL);
            video_queue_parse.set_state(Gst.State.NULL);
            video_h264_parse.set_state(Gst.State.NULL);
            video_queue_gtk.set_state(Gst.State.NULL);
            video_h264_decoder.set_state(Gst.State.NULL);
            video_queue_convert.set_state(Gst.State.NULL);

            video_queue_parse.unlink_many(video_h264_parse,
                                          video_queue_gtk,
                                          video_h264_decoder,
                                          video_queue_convert,
                                          video_convert_gtk);

            pipeline.remove_many(video_src,
                                 video_filter,
                                 video_queue_parse,
                                 video_h264_parse,
                                 video_queue_gtk,
                                 video_h264_decoder,
                                 video_queue_convert);

            video_source = new_video_source;
            video_setup = video_source.setups.first();
            new_video_source = null;

            video_src = video_source.create_element("video-src");
            video_filter = Gst.ElementFactory.make(
                "capsfilter", "video-filter");
            var fr = video_setup.framerates.first();
            var caps = new Gst.Caps.empty_simple(video_setup.media_type);
            caps.set_simple("framerate", typeof(Gst.Fraction),
                            fr.numerator, fr.denominator,
                            "width", typeof(int), video_setup.width,
                            "height", typeof(int), video_setup.height);
            video_filter.set("caps", caps);
            video_queue_gtk = Gst.ElementFactory.make(
                "queue", "video-queue-gtk");

            pipeline.add_many(video_src,
                              video_filter,
                              video_queue_gtk);
            video_src.link_many(video_filter,
                                video_queue_gtk,
                                video_convert_gtk);

            video_src.set_state(Gst.State.PLAYING);
            video_filter.set_state(Gst.State.PLAYING);

            return Gst.PadProbeReturn.DROP;
        }

        /* Gets the final destination for the video. */
        private string get_destination() {
            var now = new GLib.DateTime.now_local();
            string sts = "%04d-%02d-%02d %02d:%02d:%02d".printf(
                now.get_year(), now.get_month(), now.get_day_of_month(),
                now.get_hour(), now.get_minute(), now.get_second());
            var video_dir = GLib.Environment.get_user_special_dir(
                GLib.UserDirectory.VIDEOS);
            return string.join(GLib.Path.DIR_SEPARATOR_S, video_dir,
                               "Queso-%s.mp4".printf(sts));
        }

        /* Pushes the EOS event. */
        private void* push_eos_thread() {
            var pad = file_mp4mux.get_static_pad("src");
            var peer = pad.get_peer();
            pipeline.message_forward = true;
            peer.send_event(new Gst.Event.eos());
            return null;
        }

        /* Moves a file. */
        private void move_file(string src_path, string dest_path) {
            GLib.File src = GLib.File.new_for_path(src_path);
            GLib.File dest = GLib.File.new_for_path(dest_path);

            try {
                src.move(dest, GLib.FileCopyFlags.NONE, null);
            } catch (GLib.Error e) {
                GLib.warning("Could not move “%s” to “%s”: %s",
                             src_path, dest_path, e.message);
            }
        }

        /* Update record counter idle callback. */
        private bool update_saving_counter() {
            if (!saving) {
                video_text_record.set("text", "");
                return GLib.Source.REMOVE;
            }
            double ts = Timestamp.get_timestamp();
            if (ts - last_saving_ts < 1.0)
                return GLib.Source.CONTINUE;
            last_saving_ts = GLib.Math.floor(ts);
            if (waiting_keyframe) {
                dot_counter = dot_counter < 3 ? dot_counter + 1 : 0;
                var text = "Please wait for keyframe";
                for (int i = 0; i < dot_counter; i++)
                    text += ".";
                video_text_record.set("text", text);
                return GLib.Source.CONTINUE;
            }
            Gst.State state = Gst.State.NULL;
            Gst.State pending;
            pipeline.get_state(out state, out pending, 100);
            if (state != Gst.State.PLAYING)
                return GLib.Source.CONTINUE;
            string sts = Timestamp.to_string(ts - start_saving_ts);
            video_text_record.set("text", "•REC " + sts);
            return GLib.Source.CONTINUE;
        }

        /* Gets a random filename. */
        private string random_filename() {
            int ts = (int)Timestamp.get_timestamp();
            return string.join(GLib.Path.DIR_SEPARATOR_S,
                               GLib.Environment.get_tmp_dir(),
                               "queso-%08X.mp4".printf(ts));
        }

        private string get_info() {
            var r = new StringBuilder();
            r.append("Video source: ");
            r.append(video_source.name);
            r.append("Video setup: ");
            r.append(video_setup.to_string());
            return r.str;
        }

        [PrintfFormat]
        private void add_message(string format, ...) {
            var list = va_list();
            var message = format.vprintf(list);
            messages.offer(new Message(message));
        }

        private bool handle_message_queue() {
            if (!queuing_messages)
                return GLib.Source.REMOVE;
            var ts = Timestamp.get_timestamp();
            if (messages.is_empty || ts - message_queue_ts < 0.5)
                return GLib.Source.CONTINUE;
            if (ts - messages.peek().timestamp > 3.0)
                messages.poll();
            var m = "";
            foreach (var message in messages)
                m += message.text + "\n";
            video_text_message.set("text", m);
            message_queue_ts = ts;
            return GLib.Source.CONTINUE;
        }

        // private void try_link(Gst.Element src, Gst.Element sink) {
        //     if (!src.link(sink)) {
        //         stderr.printf("%s -> %s failed!\n",
        //                       src.name, sink.name);
        //         debug();
        //     }
        // }

        // private void link_pads(Gst.Element src, Gst.Element sink) {
        //     string a = src.name;
        //     string b = sink.name;
        //     var src_pad = src.get_static_pad("src");
        //     var sink_pad = sink.get_static_pad("sink");
        //     switch (src_pad.link(sink_pad)) {
        //     case OK:
        //         stderr.printf("%s -> %s: %s!\n", a, b, "OK");
        //         break;
        //     case WRONG_HIERARCHY:
        //         stderr.printf("%s -> %s: %s!\n", a, b, "WRONG_HIERARCHY");
        //         break;
        //     case WAS_LINKED:
        //         stderr.printf("%s -> %s: %s!\n", a, b, "WAS_LINKED");
        //         break;
        //     case WRONG_DIRECTION:
        //         stderr.printf("%s -> %s: %s!\n", a, b, "WRONG_DIRECTION");
        //         break;
        //     case NOFORMAT:
        //         stderr.printf("%s -> %s: %s!\n", a, b, "NOFORMAT");
        //         break;
        //     case NOSCHED:
        //         stderr.printf("%s -> %s: %s!\n", a, b, "NOSCHED");
        //         break;
        //     }
        // }

        // private int count = 0;
        // private void print_flags(Gst.BufferFlags flags) {
        //     var s = new StringBuilder();
        //     if ((flags & Gst.BufferFlags.LIVE) != 0)
        //         s.append("LIVE|");
        //     if ((flags & Gst.BufferFlags.DECODE_ONLY) != 0)
        //         s.append("DECODE_ONLY|");
        //     if ((flags & Gst.BufferFlags.DISCONT) != 0)
        //         s.append("DISCONT|");
        //     if ((flags & Gst.BufferFlags.RESYNC) != 0)
        //         s.append("RESYNC|");
        //     if ((flags & Gst.BufferFlags.CORRUPTED) != 0)
        //         s.append("CORRUPTED|");
        //     if ((flags & Gst.BufferFlags.MARKER) != 0)
        //         s.append("MARKER|");
        //     if ((flags & Gst.BufferFlags.HEADER) != 0)
        //         s.append("HEADER|");
        //     if ((flags & Gst.BufferFlags.GAP) != 0)
        //         s.append("GAP|");
        //     if ((flags & Gst.BufferFlags.DROPPABLE) != 0)
        //         s.append("DROPPABLE|");
        //     if ((flags & Gst.BufferFlags.DELTA_UNIT) != 0)
        //         s.append("DELTA_UNIT|");
        //     if ((flags & Gst.BufferFlags.TAG_MEMORY) != 0)
        //         s.append("TAG_MEMORY|");
        //     if ((flags & Gst.BufferFlags.SYNC_AFTER) != 0)
        //         s.append("SYNC_AFTER|");
        //     if ((flags & Gst.BufferFlags.NON_DROPPABLE) != 0)
        //         s.append("NON_DROPPABLE|");
        //     if ((flags & Gst.BufferFlags.LAST) != 0)
        //         s.append("LAST|");
        //     s.truncate(s.len-1);
        //     stderr.printf("FLAGS %d: %s\n", count++, s.str);
        // }
    }
}
