/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Namespace for timestamp functions.
     */
    namespace Timestamp {

        /* One billion double. */
        private const double ONE_BILLION_D = 1000000000.0;
        /* One billion integer. */
        private const int ONE_BILLION = 1000000000;

        /**
         * Returns the timestamp of a file.
         * @param filename the filename of the file.
         * @return the timestamp of a file, or the current timestamp if there is
         *         an error.
         */
        public static double get_file_timestamp(string filename) {
            try {
                var file = GLib.File.new_for_path(filename);
                var info = file.query_info("time::modified,time::modified-usec",
                                           GLib.FileQueryInfoFlags.NONE);
                var dt = info.get_modification_date_time();
                if (dt == null) {
                    return get_timestamp();
                }
                int64 ts = dt.to_unix();
                double usecs = dt.get_seconds();
                int secs = (int)usecs;
                return ts + (usecs - secs);
            } catch (GLib.Error e) {
                GLib.warning("There was an error reading from ‘%s’.\n", filename);
            }
            return get_timestamp();
        }

        /**
         * Returns the current timestamp.
         * @return the current timestamp.
         */
        public static double get_timestamp() {
            Posix.timespec current;
            Posix.clock_gettime(Posix.CLOCK_REALTIME, out current);

            long ts = current.tv_sec * ONE_BILLION + current.tv_nsec;
            return ts / ONE_BILLION_D;
        }

        /**
         * Returns a string representation of a timestamp.
         * @param timestamp the timestamp.
         * @param show_millis whether to show milliseconds.
         * @return a string representation of a timestamp.
         */
        public static string to_string(double timestamp,
                                       bool show_millis = false) {
            int tsecs = (int)timestamp;
            int hours = 0;
            int mins = tsecs / 60;
            if (mins > 60)
                hours = mins / 60;
            int secs = tsecs % 60;
            int millis = (int)((timestamp - tsecs) * 1000);
            if (show_millis)
                return "%02d:%02d:%02d.%03d".printf(hours, mins, secs, millis);
            else
                return "%02d:%02d:%02d".printf(hours, mins, secs);
        }
    }
}
