/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    public class V4LControlWidget : GLib.Object {

        private unowned V4LControl v4lctl;
        public Gtk.Label label { public get; private set; }
        public Gtk.Widget widget { public get; private set; }

        private int64[] valid_values;

        public V4LControlWidget(V4LControl v4lctl) {
            this.v4lctl = v4lctl;
            label = new Gtk.Label("<b>" + v4lctl.name + "</b>");
            label.use_markup = true;
            label.xalign = 0.0f;
            label.halign = Gtk.Align.START;
                switch (v4lctl.v4l_type) {
                case V4LControlType.INT:
                    widget = get_integer_widget();
                    break;
                case V4LControlType.BOOLEAN:
                    widget = get_boolean_widget();
                    break;
                case V4LControlType.MENU:
                    widget = get_menu_widget();
                    break;
                default:
                    widget = new Gtk.Label("Not implemented");
                    return;
                }
        }

        private Gtk.Widget get_menu_widget() {
            var combo = new Gtk.ComboBoxText();
            int a = -1, c = 0;
            var str = "";
            for (int i = 0; i < v4lctl.menu_size; i++) {
                unowned V4LControlMenuItem mi;
                mi = v4lctl.nth_menu_item(i);
                if (mi.disabled)
                    continue;
                combo.append_text(mi.name);
                if (mi.index == v4lctl.value) {
                    a = c;
                    str = mi.name;
                }
                c++;
            }
            combo.active = a;
            combo.active_id = str;
            combo.changed.connect(menu_widget_changed);
            return combo;
        }

        private void menu_widget_changed() {
            var combo = widget as Gtk.ComboBoxText;
            for (int i = 0; i < v4lctl.menu_size; i++) {
                unowned V4LControlMenuItem mi;
                mi = v4lctl.nth_menu_item(i);
                if (mi.disabled)
                    continue;
                if (combo.get_active_text() == mi.name)
                    v4lctl.value = mi.index;
            }
        }

        private Gtk.Widget get_boolean_widget() {
            var check = new Gtk.CheckButton();
            check.active = v4lctl.value != 0;
            check.toggled.connect(check_toggled);
            return check;
        }

        private void check_toggled() {
            var check = widget as Gtk.ToggleButton;
            var value = check.active ? 1 : 0;
            if (v4lctl.value == value)
                return;
            v4lctl.value = value;
            if (v4lctl.value != value)
                check.active = value == 0;
        }

        private Gtk.Widget get_integer_widget() {
            var scale = new Gtk.Scale.with_range(Gtk.Orientation.HORIZONTAL,
                                                 v4lctl.minimum,
                                                 v4lctl.maximum,
                                                 v4lctl.step);
            scale.set_value(v4lctl.value);
            scale.draw_value = false;
            scale.hexpand = true;
            scale.vexpand = false;
            scale.value_changed.connect(round_scale);
            var adj = scale.adjustment;
            adj.step_increment = adj.page_increment = v4lctl.step;
            var vv = new Gee.ArrayList<int64?>();
            for (int64 i = v4lctl.minimum; i <= v4lctl.maximum;
                 i += (int64)v4lctl.step)
                vv.add(i);
            int i = 0;
            valid_values = new int64[vv.size];
            foreach (var v in vv) {
                valid_values[i++] = v;
            }
            return scale;
        }

        private int64 closest(int64 val, int64 min, int64 max) {
            if (min >= max)
                return valid_values[min];
            int64 med = min + (max - min) / 2;
            if (val == valid_values[med])
                return val;
            if (val < valid_values[med])
                return closest(val, min, med - 1);
            return closest(val, med + 1, max);
        }

        private void round_scale() {
            var scale = widget as Gtk.Scale;
            int64 val = (int64)GLib.Math.round(scale.get_value());
            val = closest(val, 0, valid_values.length-1);
            scale.set_value(val);
            if (v4lctl.value == val)
                return;
            v4lctl.value = val;
            if (v4lctl.value != val)
                scale.set_value(v4lctl.value);
        }
    }
}
