/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for video setups.
     */
    public class VideoSetup : GLib.Object, Gee.Comparable<VideoSetup> {

        /**
         * Video raw format.
         */
        public const string RAW = "video/x-raw";

        /**
         * Video H264 format.
         */
        public const string H264 = "video/x-h264";

        /**
         * The media type.
         */
        public string media_type { public get; private set; }

        /**
         * The format.
         */
        public string format { public get; private set; }

        /**
         * The width.
         */
        public int width { public get; private set; }

        /**
         * The height.
         */
        public int height { public get; private set; }

        /**
         * The framerates.
         */
        public Gee.TreeSet<Framerate> framerates { public get; private set; }

        /**
         * The area.
         */
        public int area { public get { return width * height; } }

        /**
         * Whether is landscape.
         */
        public bool is_landscape { public get { return width > height; } }

        /**
         * Constructs a video setup.
         */
        public VideoSetup(string media_type, string? format,
                          int width, int height,
                          Gee.TreeSet<Framerate> framerates) {
            this.media_type = media_type;
            this.format = format;
            this.width = width;
            this.height = height;
            this.framerates = framerates;
        }

        /**
         * Returns a string representation of the video setup.
         * @return a string representation of the video setup.
         */
         public string to_string() {
            string f = (format == null) ? "" :
                "format=(string)%s, ".printf(format);
            string sfr = "";
            if (framerates.size == 1) {
                var fr = framerates.first();
                sfr = "framerate=(fraction)%d/%d".printf(fr.numerator,
                                                         fr.denominator);
            } else {
                foreach (var fr in framerates) {
                    if (sfr == "") {
                        sfr = "framerate=(fraction){ %d/%d".printf(
                            fr.numerator,fr.denominator);
                    } else {
                        sfr += ", %d/%d".printf(fr.numerator, fr.denominator);
                    }
                }
                sfr += " }";
            }

            return "%s, %swidth=(int)%d, height=(int)%d, %s".printf(
                            media_type, f, width, height, sfr);
        }

        /**
         * Compares the two video setups.
         * @param setup the video setup to compare with.
         * @return -1 if this setup it's better than the one received; 0 if
         *         they are equal; 1 if it's worse.
         */
        public int compare_to(VideoSetup vs) {
            if (this.media_type == H264 && vs.media_type != H264)
                return -1;
            if (this.media_type != H264 && vs.media_type == H264)
                return 1;
            var tfr = this.framerates.first();
            var vfr = vs.framerates.first();
            if (tfr.compare_to(vfr) < 0)
                return -1;
            if (tfr.compare_to(vfr) > 0)
                return 1;
            if (this.is_landscape && !vs.is_landscape)
                return -1;
            if (!this.is_landscape && vs.is_landscape)
                return 1;
            return vs.area - this.area;
        }
    }
}
