/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for video source settings dialogs.
     */
    [GtkTemplate (ui = "/mx/unam/Queso/video-source-settings-dialog.ui")]
    public class VideoSourceSettingsDialog : Gtk.Dialog {

        private VideoSources video_sources;
        private VideoSource videodev;
        private V4LControlManager v4lmgr;
        private Gee.ArrayList<V4LControlWidget> v4lwidgets;

        [GtkChild]
        private Gtk.ComboBoxText video_source_combo;

        [GtkChild]
        private Gtk.Grid vs_grid;

        public VideoSourceSettingsDialog(ApplicationWindow window,
                                         VideoSources video_sources) {
            GLib.Object(use_header_bar: 1);
            transient_for = window;
            this.video_sources = video_sources;
            video_sources.notify["video-source"].connect(video_source_changed);

            if (video_sources == null)
                return;
            foreach (var device in video_sources.sources) {
                video_source_combo.append_text(device.name);
            }
            video_source_combo.active_id = video_sources.video_source.name;
            v4lwidgets = new Gee.ArrayList<V4LControlWidget>();

            set_device(video_sources.video_source);
        }

        private void set_device(VideoSource videodev) {
            if (videodev == this.videodev)
                return;
            this.videodev = videodev;
            if (video_source_combo.active_id != videodev.name)
                video_source_combo.active_id = videodev.name;
            var dp = videodev.device_path;
            if (dp == VideoSources.DEV_NULL) {
                var label = new Gtk.Label("No settings available for\n" +
                                          videodev.name);
                vs_grid.attach(label, 0, 0, 1, 1);
                label.margin = 48;
                label.visible = true;
                return;
            }

            int row = 0;
            v4lmgr = new V4LControlManager(dp);
            for (int i = 0; i < v4lmgr.size; i++) {
                unowned V4LControl v4lctrl = v4lmgr.nth_control(i);
                var v4lwidget = new V4LControlWidget(v4lctrl);
                v4lwidgets.add(v4lwidget);
                vs_grid.attach(v4lwidget.label, 0, row, 2, 1);
                vs_grid.attach(v4lwidget.widget, 2, row++, 2, 1);
                v4lwidget.label.visible = v4lwidget.widget.visible = true;
            }
        }

        [GtkCallback]
        private void on_close_clicked() {
            this.close();
        }

        /* The on key press event callback. */
        [GtkCallback]
        private bool on_key_press_event(Gdk.EventKey event) {
            uint event_key;
            Gdk.keyval_convert_case(event.keyval, null, out event_key);
            var default_modifiers = Gtk.accelerator_get_default_mod_mask();
            if (event_key == Gdk.Key.W &&
                (event.state & default_modifiers) ==
                Gdk.ModifierType.CONTROL_MASK) {
                this.close();
                return true;
            }
            return false;
        }

        [GtkCallback]
        private void on_video_source_combo_changed() {
            video_sources.update_video_source(video_source_combo.active_id);
        }

        private void video_source_changed() {
            if (video_sources.video_source == this.videodev)
                return;
            vs_grid.foreach(c => vs_grid.remove(c));
            v4lwidgets.clear();
            v4lmgr = null;
            set_device(video_sources.video_source);
        }
    }
}
