/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class to manage devices in the system.
     */
    public class VideoSources : GLib.Object {

        /**
         * Null device.
         */
        public const string DEV_NULL = "/dev/null";

        /**
         * The video devices.
         */
        public Gee.TreeSet<VideoSource> sources { public get; private set; }

        /**
         * The selected video device.
         */
        public VideoSource video_source { public get; private set; }

        private Gee.Iterator<VideoSource> vd_iter;

        /* The device monitor. */
        private Gst.DeviceMonitor monitor;
        /* The bus. */
        private Gst.Bus bus;

        /* Constructs a device manager. */
        construct {
            sources = new Gee.TreeSet<VideoSource>();
            sources.add(new VideoSource.test());
            monitor = new Gst.DeviceMonitor();
            bus = monitor.get_bus();
            bus.add_watch(GLib.Priority.DEFAULT, monitor_devices);

            monitor.add_filter("Video/Source",
                               new Gst.Caps.empty_simple(VideoSetup.RAW));
            monitor.add_filter("Video/Source",
                               new Gst.Caps.empty_simple(VideoSetup.H264));

            monitor.start();

            var devices = monitor.get_devices();
            unowned GLib.List<Gst.Device> t = devices;
            while (t != null) {
                switch (t.data.get_device_class()) {
                case "Video/Source":
                    sources.add(new VideoSource(t.data));
                    break;
                }
                t = t.next;
            }

            vd_iter = sources.iterator();
            if (vd_iter.has_next()) {
                vd_iter.next();
                video_source = vd_iter.get();
            }
        }

        public void next_video_source() {
            if (!vd_iter.has_next())
                vd_iter = sources.iterator();
            vd_iter.next();
            video_source = vd_iter.get();
        }

        public void update_video_source(string name) {
            if (video_source.name == name)
                return;
            foreach (var device in sources) {
                if (device.name == name)
                    video_source = device;
            }
        }

        /* Monitors the devices. */
        private bool monitor_devices(Gst.Bus bus, Gst.Message message) {
            Gst.Device device;
            string name;

            switch (message.type) {
            case Gst.MessageType.DEVICE_ADDED:
                message.parse_device_added(out device);
                name = device.get_display_name();
                stdout.printf("Device added: %s\n", name);
                break;
            case Gst.MessageType.DEVICE_REMOVED:
                message.parse_device_removed(out device);
                name = device.get_display_name();
                stdout.printf("Device added: %s\n", name);
                break;
            default:
                break;
            }

            return GLib.Source.CONTINUE;
        }
    }
}
